from __future__ import annotations
from typing import Dict
from dataclasses import dataclass

from .type_map import TYPE_MAP


@dataclass(frozen=True)
class ColumnInfo:
    description: str
    mode: str
    type: str
    name: str

    @classmethod
    def create(cls, d: Dict[str, str]) -> ColumnInfo:
        description = d["description"] if "description" in d else ""
        mode = d["mode"] if "mode" in d else "nullable"
        type = d["type"]
        name = d["name"]
        return cls(description, mode, type, name)

    def typehint(self) -> str:
        name = TYPE_MAP[self.type.lower()].__name__
        return f"{self.name}: Union[{name}, Sequence[{name}]]"



from typing import Dict, Optional, Type, Union
from pathlib import Path
import re
import json
import glob
import os
import subprocess as sp
from io import StringIO
import logging


from .table_info import TableInfo
from .column_info import ColumnInfo
from .table import Table

LOGGER = logging.getLogger(__name__)


def snake_to_camel(s: str) -> str:
    return "".join([x.title() for x in s.split("_")])


def load(s: Union[str, Path]) -> Type[Table]:
    path = Path(s)
    file = path.as_posix().split("/")[-1]
    body = re.sub(r".json$", "", file)
    name = snake_to_camel(body)
    with open(path) as fp:
        content = "".join(fp.readlines())
    return loads(content, name)


def load_bq(project: Optional[str], dataset: str, table: str) -> Type[Table]:
    cmd = [
        "bq", "show"
    ]
    if project is not None:
        cmd += [f"--project_id={project}"]
    cmd += [
        "--schema",
        f"--format=prettyjson",
        f"{dataset}.{table}"
    ]
    ret = sp.run(
        cmd,
        text=True,
        stdout=sp.PIPE,
        stderr=sp.PIPE
    )
    if ret.returncode != 0:
        raise RuntimeError(
            f"command: {' '.join(cmd)} failed. stdout={ret.stdout}, stderr={ret.stderr}")
    return loads(ret.stdout, snake_to_camel(table))


def loads(s: str, name: str) -> Type[Table]:
    obj = json.loads(s)
    cols = [
        ColumnInfo.create(d) for d in obj
    ]
    nulls = {
        col.name: None for col in cols
        if col.mode.upper() == "NULLABLE"
    }

    tinfo = TableInfo(cols)
    return type(name, (Table,), {
        "_table_info": tinfo,
        **nulls
    })


def load_dir(json_dir: Union[str, Path]) -> Dict[str, Type[Table]]:
    ls = [load(js) for js in glob.glob(os.path.join(json_dir, "**.json"))]
    return {
        t.__name__: t for t in ls
    }



from dataclasses import dataclass
from typing import Iterable, Optional
import glob
import os

import click

from .table import Table, typehint
from .load import load, load_bq


@click.group()
def main():
    pass


def print_typehint(tables: Iterable[Table]):
    ret = """from typing import Sequence, Union
from datetime import date, datetime, time, timedelta

from bqsc import Table
"""

    for t in tables:
        ret += typehint(t) + "\n" * 2
    print(ret)


@main.command()
@click.argument("json_dir", type=click.Path(file_okay=False, readable=True))
def dir(json_dir: str):
    if not os.path.isdir(json_dir):
        raise ValueError(f"{json_dir} is not a directory")
    print_typehint([
        load(js)()
        for js in glob.glob(os.path.join(json_dir, "**.json"))
    ])


@dataclass
class TableRepr:
    project: Optional[str]
    dataset: str
    table: str


def parse_table(s: str) -> TableRepr:
    pos = s.find(":")
    if pos > 0:
        proj = s[:pos]
        s = s[(pos + 1):]
    else:
        proj = None
    pos = s.find(".")
    if pos < 1:
        raise ValueError(
            f"table format should be [project:]dataset.table: given value = {s}")
    ds = s[:pos]
    table = s[(pos+1):]
    return TableRepr(proj, ds, table)


@main.command()
@click.argument("table", type=str)
def bq(table: str):
    rep = parse_table(table)
    ty = load_bq(rep.project, rep.dataset, rep.table)()
    print_typehint([ty])


if __name__ == "__main__":
    main()



from typing import Iterable, Any

from .column_info import ColumnInfo
from .type_map import TYPE_MAP


class TableInfo:
    def __init__(self, cinfos: Iterable[ColumnInfo]) -> None:
        self.column_infos = list(cinfos)
        self.column_names = [c.name for c in self.column_infos]
        self.column_types = {
            c.name: TYPE_MAP[c.type.lower()] for c in self.column_infos
        }

    def type_check(self, name: str, value: Any) -> bool:
        ty = self.column_types[name]
        return isinstance(value, ty)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bqsc']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.3,<9.0.0', 'mypy>=0.910,<0.911', 'pandas>=1.3.4,<2.0.0']

entry_points = \
{'console_scripts': ['bqsc = bqsc.cli:main']}

setup_kwargs = {
    'name': 'bqsc',
    'version': '3.0.3',
    'description': 'Define schema object from bigquery schema definition json file',
    'long_description': None,
    'author': 'Yasunori Horikoshi',
    'author_email': 'hotoku@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Iterable

import numpy as np
import pandas as pd

from .table_info import TableInfo


class NotDefinedColumn(Exception):
    def __init__(self, name: str) -> None:
        super().__init__(f"{name} is not defined.")


class TypeMismatch(Exception):
    def __init__(self, name: str, value: Any, expected: type) -> None:
        super().__init__(
            f"column: {name}, expected: {expected}, given: {value}")


class Table:
    _table_info: TableInfo

    def __setattr__(self, name: str, value: Any) -> None:
        if name not in self._table_info.column_names:
            raise NotDefinedColumn(name)

        if isinstance(value, Iterable):
            it = iter(value)
            v = next(it)
        else:
            v = value

        if not self._table_info.type_check(name, v):
            raise TypeMismatch(name, v, self._table_info.column_types[name])

        super().__setattr__(name, value)


def typehint(table: Table) -> str:
    ret = f"""
class {type(table).__name__}(Table):
"""
    for col in table._table_info.column_infos:
        ret += "    " + col.typehint() + "\n"
    ret += "    ..."
    return ret


def is_list_like(v: any) -> bool:
    if isinstance(v, str):
        return False
    else:
        return isinstance(v, Iterable)


def dataframe(table: Table) -> pd.DataFrame:
    info = table._table_info
    dic = {
        col.name: table.__getattribute__(col.name)
        for col in info.column_infos
    }
    is_scalar = [
        not is_list_like(v) for _, v in dic.items()
    ]

    if np.all(is_scalar):
        return pd.DataFrame(dic, index=[0])
    else:
        return pd.DataFrame(dic)



__version__ = "3.0.3"

from .load import (
    load as load,
    loads as loads,
    load_dir as load_dir,
    load_bq as load_bq
)
from .table import (
    Table as Table,
    TypeMismatch as TypeMismatch,
    NotDefinedColumn as NotDefinedColumn,
    dataframe as dataframe
)



from typing import Dict
from datetime import date, datetime, time, timedelta


TYPE_MAP: Dict[str, type] = {
    "string": str,
    "integer": int,
    "float": float,
    "boolean": bool,
    "date": date,
    "datetime": datetime,
    "time": time,
    "interval": timedelta
}
